(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *
 * POR_Testing --- a theory on partial order reduction for testing.
 *
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2009-15 Univ. Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)

theory  POR_Testing
imports 
   (* Testing *) 
   (*"../../../src/codegen_fsharp/Code_Integer_Fsharp"*) 
   "../../../src/Monads"
   Permutations
begin


section{* Partial Order Reduction Approach *}


text{* The continuation monad $M$ is conceptually superfluous; it is used 
       to carry the result type-parameter @{typ "'o'"} to maintain full
       generality. This can not be captured by a universal quantification \emph{inside}
       the definition. *}

definition po_is_exec_consistent :: 
           "('\<iota> \<Rightarrow> ('o,'\<sigma>)MON\<^sub>S\<^sub>E) \<Rightarrow> ('\<iota> \<Rightarrow> '\<iota> \<Rightarrow>  bool) \<Rightarrow>  ('o \<Rightarrow> 'o \<Rightarrow> ('o','\<sigma>)MON\<^sub>S\<^sub>E) \<Rightarrow> bool"
where      "po_is_exec_consistent iostep ord M =            
                (\<forall> \<iota> \<iota>'. 
                    (ord \<iota>' \<iota>) \<longrightarrow>                
                         ((o1 \<leftarrow> iostep \<iota>;  (o2 \<leftarrow> iostep \<iota>'; M o1 o2))) = 
                         ((o2 \<leftarrow> iostep \<iota>'; (o1 \<leftarrow> iostep \<iota>;  M o1 o2))))"


definition itrf_is_exec_consistent :: 
           "('\<iota> \<Rightarrow> ('o,'\<sigma>)MON\<^sub>S\<^sub>E) \<Rightarrow> ('\<iota> \<Rightarrow> '\<iota> \<Rightarrow>  bool) \<Rightarrow>  ('o \<Rightarrow> 'o \<Rightarrow> ('o','\<sigma>)MON\<^sub>S\<^sub>E) \<Rightarrow> bool"
where      "itrf_is_exec_consistent iostep itrf M =            
                (\<forall> \<iota> \<iota>'. 
                    (\<not> itrf \<iota> \<iota>') \<longrightarrow>                
                         ((o1 \<leftarrow> iostep \<iota>;  (o2 \<leftarrow> iostep \<iota>'; M o1 o2))) = 
                         ((o2 \<leftarrow> iostep \<iota>'; (o1 \<leftarrow> iostep \<iota>;  M o1 o2))))"



subsection{* An abstract notion of interference *}


locale interference_relation =
  fixes interfer :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "##" 50)
    and iostep   :: "'a \<Rightarrow> ('o,'\<sigma>)MON\<^sub>S\<^sub>E"
  assumes interfer_sym: "x ## y \<Longrightarrow> y ## x"
      and interfer_consistent: "\<And>M. itrf_is_exec_consistent iostep interfer M"
begin

(*
  (* Should be an "assumes" *)
  definition interference_is_exec_consistent ::
             "('o \<Rightarrow> 'o \<Rightarrow> ('o','\<sigma>)MON_SE) \<Rightarrow> bool"
  where      "interference_is_exec_consistent M =            
                (\<forall> a a'. 
                    (\<not> a' ## a) \<longrightarrow>                
                         ((o1 \<leftarrow> iostep a;  (o2 \<leftarrow> iostep a'; M o1 o2))) = 
                         ((o2 \<leftarrow> iostep a'; (o1 \<leftarrow> iostep a;  M o1 o2))))"
*)

lemma interfer_sym': "\<not> x ## y \<Longrightarrow> \<not> y ## x"
by(erule contrapos_nn, auto intro: interfer_sym)

end 


subsection{* Standardization algorithm
            parameterised on an interference relation and an ordering relation *}


locale standardization_order = ordering + interference_relation +
  assumes sq_not_strd: "  (x::'b) ## y \<Longrightarrow> \<not> (x \<prec> y)"
      and not_sq_strd: "\<not> x ## y \<Longrightarrow> (x \<prec> y \<or> x = y \<or> y \<prec> x)"
begin

lemma ordering_is_exec_consistent: "po_is_exec_consistent iostep (op \<prec>) M"
proof(simp add:po_is_exec_consistent_def, auto)
  assume "\<iota>' \<prec> \<iota>"
  then have "\<not> \<iota> ## \<iota>'" by (metis interfer_sym sq_not_strd)
  then have "( o1 \<leftarrow> iostep \<iota>; o2 \<leftarrow> iostep \<iota>'; M o1 o2) = ( o2 \<leftarrow> iostep \<iota>';  o1 \<leftarrow> iostep \<iota>; M o1 o2)"
  using interfer_consistent
  apply (simp only:itrf_is_exec_consistent_def)
  (* Hum... modus ponens? *)
oops


(* A variant of [not_sq_strd]. *)
lemma four_cases: "x ## y \<or> x \<prec> y \<or> x = y \<or> y \<prec> x"
by (metis not_sq_strd)


(* Based on the ordering relation \<prec>, we will introduce a standardization algorithm
  that reorders a list while preserving the relative positions of interfering
  elements. *)


(* The simpler definition of ordered: no two consecutive elements are in the
  wrong order for \<prec>. That is, an element in a non-decreasing list can be followed by:
  1/ another occurrence of the same element;
  2/ a greater element; or
  3/ an interfering element. *)
definition non_decreasing_list :: "'b list \<Rightarrow> bool"
where "non_decreasing_list S = (\<forall>i. 0 \<le> i \<longrightarrow> Suc i < length S \<longrightarrow> \<not> S!(Suc i) \<prec> S!i)"

(* The following lemmas give an inductive version of non-decreasingness. *)
lemma ndl_rec_nil [simp]: "non_decreasing_list []"
      by (simp add:non_decreasing_list_def)
lemma ndl_rec_sing [simp]: "non_decreasing_list [a]"
      by (simp add:non_decreasing_list_def)
lemma ndl_rec_cons2 [simp]: "non_decreasing_list (a#b#S) = (\<not> b \<prec> a \<and> non_decreasing_list (b#S))"
      by (simp add:non_decreasing_list_def, auto, case_tac "i", simp_all)
lemma ndl_rec_cons1: "non_decreasing_list (a#S) = (S = [] \<or> (\<not> hd S \<prec> a \<and> non_decreasing_list S))"
      by (cases "S", simp_all)


(* Characterization of "permutable elements", ie elements that can be brought
  to the head of the list. They are the elements in the list that are smaller
  than every element before them. *)

fun pe :: "'b \<Rightarrow> 'b list \<Rightarrow> bool"
where pe_nil:  "pe a []    = False"
    | pe_cons: "pe a (b#S) = (a = b \<or> (a \<prec> b \<and> pe a S))"


lemma pe_member: "pe a S \<Longrightarrow> List.member S a"
proof(induct S)
  case Nil then show ?case by simp
next
  case (Cons b S) then show ?case by(auto, simp_all add: List.member_def)
qed


(* Three following lemmas: an element [a] of a list [S] is a permutable element
  if and only if it is smaller than all the elements of [S] that occurs before
  the first occurrence of [a]. *)
lemma pe_first_occ: "pe a S \<Longrightarrow> i < first_occ a S \<Longrightarrow> a \<prec> S!i"
proof(induct S arbitrary:i)
  case Nil then show ?case by simp
next
  case (Cons b S)
  { assume "a=b"
    then have ?case using Cons
         by simp }
  moreover
  { assume "a\<noteq>b"
    then have ?case using Cons
         by (cases "i", simp_all) }
  ultimately show ?case by blast
qed

lemma first_occ_pe: "List.member S a \<Longrightarrow> (\<forall>i < first_occ a S. a \<prec> S!i) \<Longrightarrow> pe a S"
proof(induct S)
  case Nil then show ?case by (simp add:List.member_def)
next
  case (Cons b S)
  { assume "a=b"
    then have ?case using Cons by simp }
  moreover
  { assume "a\<noteq>b"
    then have ?case using Cons
         apply(simp)
         apply(rule conjI)
          apply(metis nth_Cons_0 zero_less_Suc)
         by(metis less_Suc_eq_0_disj member_rec(1) nth_Cons_Suc) }
  ultimately show ?case by blast
qed

lemma member_not_pe_implies_exists_smaller:
  "List.member S a \<Longrightarrow> \<not> pe a S \<Longrightarrow> \<exists>i<first_occ a S. \<not> a \<prec> S!i"
by (metis first_occ_pe)


(* Additional lemmas on the interaction between map, upt, take, and drop. *)
lemma take_upt_from0: "i \<le> n \<Longrightarrow> [0..<i] = take i [0..<n]"
by (metis plus_nat.add_0 take_upt)

lemma drop_upt_from0: "[i..<n] = drop i [0..<n]"
by (metis plus_nat.add_0 drop_upt)

lemma take_map_upt: "i \<le> length S \<Longrightarrow> take i S = map (op ! S) [0..<i]"
by (metis map_nth take_map take_upt_from0)

lemma drop_map_upt: "drop i S = map (op ! S) [i..<length S]"
by (metis map_nth drop_map drop_upt_from0)


(* Crucial lemma: bringing any permutable element to the head of the list preserves
  the relative order of interfering elements. *)

lemma permuting_pe_preserves_interferences:
  "pe a S \<Longrightarrow>
   let U = (a # (remove1 a S)) in
   \<exists>ip. is_perm [0 ..<length S] ip
        \<and> (U = map (op ! S) ip)
        \<and> (\<forall>k<length S. \<forall>j<k. U!j ## U!k \<longrightarrow> ip!j < ip!k)"
proof -
  (* Let [a] be a permutable element of [S], and [i] the index of its first occurrence. *)
  assume "pe a S"
  then have a_member: "List.member S a"
       by (metis pe_member)
  def  i \<equiv> "first_occ a S"
  have i_reverse: "first_occ a S = i" by (simp add:i_def)
  have i_lt_length: "i < length S" using i_def a_member
       by (metis first_occ_member_length)
  
  (* Write [T] for the result of removing [a] from [S]. *)
  def  T \<equiv> "remove1 a S"
  then have T_length: "length S = Suc (length T)" using T_def a_member
       by (metis is_perm_implies_same_length is_perm_refl is_perm_remove1_front length_Suc_conv)
  have uptT_length: "length T = length [0..<length T]"
       by simp
  def  U \<equiv> "a # (remove1 a S)"

  (* Define [ip] as the permutation bringing the [i]th element at the head.
    This is our candidate index permutation. *)
  def ip \<equiv> "ext_perm i [0..<length T]"
  have ip_alt: "ip = i # [0..<i] @ [Suc i..<length S]"
       by (metis (hide_lams, no_types) T_length ext_perm_def i_lt_length ip_def le0 le_antisym less_imp_le_nat less_not_refl lift_from_upt upt_Suc upt_eq_Nil_conv)

  (* First part: [ip] is a permutation of the indexes of [S]. *)
  have *: "is_perm [0 ..<length S] ip"
       apply (simp add:ip_def T_length del:upt_Suc, subst uptT_length)
       apply (rule ext_perm_correct[THEN is_perm_sym])
         apply simp
        apply (metis T_length i_lt_length le_less_linear not_less_eq uptT_length)
       apply simp
       done

  (* Second part: [ip] relates [S] and [U]. *)
  then have **: "U = map (op ! S) ip" using a_member i_lt_length
       (* Putting [auto] in this line or on a different line does not give the same goals *)
       apply (simp add:U_def T_def ip_alt remove1_first_occ i_def)
       apply auto
         apply (metis first_occ_member)
        apply (metis le_eq_less_or_eq take_map_upt)
       apply (metis drop_map_upt)
       done

  (* Third part: [ip] preserves the relative order of interfering elements. *)
  { fix j k
    assume a_pe:     "pe a S"
       and k_range:  "k < length S"
       and j_range:  "j < k"
       and jk_inter: "U!j ## U!k"
    then have "ip ! j < ip ! k"
    proof -
    (* Proof is by case on [j], the first index considered.
       1. If [j] is [0], then we are looking at the element [a] that has been permuted.
          This is the interesting case, and it is split in two:
          1.a) If the second index considered [k] is less than [i] (the index of the
               first occurrence of [a]), then the elements have been permuted, but using
               the assumption that [a] is a permutable element we know that they are
               not interfering.
          1.b) If [k] is greater than [i], then the relative order did not change.
       2. If [j] is not [0], then the relative order did not change. *)
    (* Case 1. *)
    { assume j0: "j = 0"
      (* Case 1.a) *)
      { assume "k \<le> i"
        then have ipk_lt_i: "ip!k < i" using i_lt_length T_length j_range
             apply (simp add:ip_def T_length j0 ext_perm_def)
             apply (subst lift_from_def)
             apply (split split_if)
             by (metis Suc_pred le_eq_less_or_eq less_Suc_eq)
        then have "S ! (ip ! k) \<prec> a" using i_def a_member
             by (metis "**" Suc_length_conv T_def T_length U_def a_pe j0 jk_inter k_range length_map nth_Cons_0 nth_map pe_first_occ sq_not_strd)
        then have "U!k \<prec> a"
             by (metis ipk_lt_i a_pe asym i_def pe_first_occ)
        then have False
             by (metis j0 U_def interfer_sym jk_inter nth_Cons_0 sq_not_strd)
        then have ?thesis by simp }
      moreover
      (* Case 1.b) *)
      { assume ik: "i < k"
        then have "ip!k = k" using k_range
             apply (simp add:ip_def T_length j0 ext_perm_def)
             apply (subst lift_from_def)
             by (metis j0 Suc_pred j_range not_less_eq)
        then have ?thesis using ik
             by (metis j0 ip_alt nth_Cons_0) }
      (* Conclusion on case 1. *)
      ultimately have ?thesis by (metis leI) }
    moreover
    (* Case 2. *)
    { assume jn0: "0 < j"
      obtain j' where rj: "j = Suc j'" using jn0
             by (metis Suc_pred')
      obtain k' where rk: "k = Suc k'" using jn0 j_range
             by (metis less_nat_zero_code nat.exhaust)
      then have ?thesis using j_range k_range
           by (simp add:ip_def rj rk ext_perm_def T_length lift_from_mono) }
    (* Conclusion on the third part *)
    ultimately show ?thesis by blast
    qed } note *** = this

  (* Conclusion: [ip] is a valid candidate. *)
  assume "pe a S"
  then show ?thesis using U_def * ** *** by metis

qed


(* To avoid unnecessary permutations, the algorithm will always select the rightmost
  permutable element. Here is an abstract definition. A function is provided below. *)

definition rightmost_pe :: "'b \<Rightarrow> 'b list \<Rightarrow> bool"
where "rightmost_pe a S = (pe a S \<and> (\<forall>b. pe b S \<longrightarrow> first_occ b S \<le> first_occ a S))"


(* Two lemmas for enabling recursive reasoning. *)
lemma rightmost_pe_rec_greater:
  "rightmost_pe a S \<Longrightarrow> a \<prec> b \<Longrightarrow> rightmost_pe a (b#S)"
by(simp add:rightmost_pe_def, metis irrefl)

lemma rightmost_pe_rec_not_greater:
  "rightmost_pe a S \<Longrightarrow> \<not> a \<prec> b \<Longrightarrow> rightmost_pe b (b#S)"
apply (simp add:rightmost_pe_def)
by (metis dual_order.antisym first_occ_member le_less_linear pe_first_occ pe_member strict_implies_order strict_trans2)


(* This function is only defined for a non-empty list.
  To state a property on [rmpe S], one cannot compare [rmpe S] to [undefined],
  because [undefined] is an arbitrary value, that is not necessarily different
  from the actual value were are dealing with. Good practice is to instead take
  [ S \<noteq> [] ] as assumption in each lemma. *)
fun rmpe :: "'b list \<Rightarrow> 'b"
where rmpe_cons: "rmpe (a#S) = (if S = [] then a
                                else if rmpe S \<prec> a then rmpe S else a)"


(* Lemma for recursive reasoning. *)
lemma rmpe_rec: "(rmpe (a#S) = a) \<or> (rmpe (a#S) = rmpe S \<and> rmpe (a#S) \<prec> a)"
proof(cases "S")
  case Nil then show ?thesis by simp
next
  case (Cons b S') 
  { assume "rmpe S \<prec> a"
    then have ?thesis using Cons by simp }
  moreover
  { assume "\<not> rmpe S \<prec> a"
    then have ?thesis using Cons by simp }
  ultimately show ?thesis by blast
qed

(* Simple deductions (actually not used). *)
lemma rmpe_tail: "rmpe (a#S) \<prec> a \<Longrightarrow> rmpe (a#S) = rmpe S"
by (metis rmpe_rec strict_implies_not_eq)
lemma rmpe_leq: "rmpe (a#S) \<preceq> a"
by (metis rmpe_rec order_iff_strict)


(* Three following lemmas:
  1/ [rmpe] returns an element of the list
  2/ [rmpe] returns a permutable element of the list
  3/ [rmpe] returns the rightmost permutable element of the list
  The last one is not used in the subsequent development but proves the function
  correct with respect to the abstract definition of a rightmost permutable element.
*)
lemma rmpe_member: "S \<noteq> [] \<Longrightarrow> List.member S (rmpe S)"
proof(induct S)
  case Nil then show ?case by simp
next
  case (Cons a S)
  { assume "rmpe (a#S) = a"
    then have ?case by (simp add:List.member_rec) }
  moreover
  { assume "rmpe (a#S) = rmpe S"
    then have ?case using Cons
    by(cases "S", simp_all add:List.member_def) }
  ultimately show ?case by (metis rmpe_rec)
qed


lemma rmpe_pe: "S \<noteq> [] \<Longrightarrow> pe (rmpe S) S"
proof(induct S)
  case Nil then show ?case by simp
next
  case (Cons a S)
  { assume "rmpe (a#S) = a"
    then have ?case by simp }
  moreover
  { assume "rmpe (a#S) = rmpe S \<and> rmpe (a#S) \<prec> a"
    then have ?case
         apply(cases "S", simp)
         by (metis Cons.hyps list.distinct(1) local.pe_cons) }
  ultimately show ?case by (metis rmpe_rec)
qed


(* Not used, but proves the correctness of [rmpe] with respect to the abstract
  definition of rightmost permutable elements. *)
lemma rmpe_rightmost_pe: "S \<noteq> [] \<Longrightarrow> rightmost_pe (rmpe S) S"
proof(induct S)
  case Nil then show ?case by simp
next
  case (Cons a S) then show ?case 
  apply auto
     apply(simp add:rightmost_pe_def)
    apply(simp add:rightmost_pe_rec_greater)
   apply(simp add:rightmost_pe_def)
  apply(simp add:rightmost_pe_rec_not_greater)
  done
qed


(* Any element smaller than the rightmost permutable element is not permutable.
  Not used, but gives a variant of the correctness of [rmpe]. *)
lemma smaller_than_rmpe_not_pe:
  "a \<prec> rmpe S \<Longrightarrow> \<not> pe a S"
proof(cases "S")
  case Nil then show ?thesis by simp
next
  case (Cons b S')
  then have rm: "rightmost_pe (rmpe S) S"
       by (metis list.distinct(1) rmpe_rightmost_pe)
  { assume a_pe: "pe a S"
       and a_smaller: "a \<prec> rmpe S"
    then have "first_occ a S \<le> first_occ (rmpe S) S" using rm
         by (simp add: rightmost_pe_def)
    then have "first_occ a S < first_occ (rmpe S) S" using a_smaller
         by (metis a_pe eq_iff first_occ_member less_imp_le_nat linorder_cases pe_member rightmost_pe_def rm strict_iff_order)
    then have "rmpe S \<prec> S ! (first_occ a S)" using rm
         by (metis pe_first_occ rightmost_pe_def)
    then have False using a_smaller
         by (metis a_pe asym first_occ_member pe_member)
  } note * = this

  assume "a \<prec> rmpe S"
  then show ?thesis using * by blast
qed



(* Final variation on the correctness of [rmpe]. This is the one that is actually used
  in the proof of standardization (induction step of [standard_non_decreasing]). *)
lemma rmpe_rmpe_remove1:
  "length S > 1 \<Longrightarrow> \<not> rmpe (remove1 (rmpe S) S) \<prec> rmpe S"
proof(induct S)
  case Nil then show ?case by simp
next
  case (Cons a S)
  { assume "rmpe (a#S) = a"
    then have ?case using Cons
         by (simp add:irrefl split:split_if_asm) }
  moreover
  { assume "rmpe (a#S) \<prec> a"
    then have ?case using Cons
         apply(simp add:irrefl)
         by (metis asym length_Suc_conv less_Suc0 local.rmpe_cons nat.inject nat_neq_iff remove1.simps(2) rmpe.cases zero_less_Suc) }
  ultimately show ?case by (metis rmpe_rec)
qed


(* Termination argument for the standardization algorithm. *)
lemma [simp]: "rmpe S \<prec> a \<Longrightarrow> S \<noteq> [] \<Longrightarrow> rmpe S \<noteq> a \<Longrightarrow> length (remove1 (rmpe S) S) < length S"
      apply(simp only:length_remove1 split:split_if)
      by (metis diff_less in_set_member length_greater_0_conv rmpe_member zero_less_one)


(* This standardization algorithm selects the rightmost permutable element and
  brings it at the head of the list. Then it goes on with what remains. *)

fun standard :: "'b list \<Rightarrow> 'b list"
where standard_nil:  "standard []    = []"
    | standard_cons: "standard (a#S) = (let b = rmpe (a#S) in
                                        b # standard (remove1 b (a#S)))"


(* Lemmas for recursive reasoning. *)
lemma standard_not_nil:
  "S \<noteq> [] \<Longrightarrow> standard S = rmpe S # (standard (remove1 (rmpe S) S))"
proof(induct "S")
  case Nil then show ?case by simp
next
  case (Cons a S) then show ?case by (simp only:standard_cons Let_def)
qed

lemma standard_not_nil_adhoc:
  "Suc n = length S \<Longrightarrow> standard S = rmpe S # (standard (remove1 (rmpe S) S))"
by (metis Suc_length_conv list.distinct(1) standard_not_nil)


(* Correctness, part I: standardization gives a permutation of the list. *)

lemma standard_is_perm : "is_perm (standard S) S"
proof(induct "length S" arbitrary:S)
  case 0 then show ?case by simp
next
  case (Suc n) then show ?case
  proof(cases "S")
  case Nil then show ?thesis by simp
  next
  case (Cons a S')
  then have rm: "List.member S (rmpe S)"
       by (metis rmpe_member member_rec)
  then show ?thesis using Cons Suc.hyps
       apply(simp only:Cons standard_cons Let_def)
       by (metis Cons is_perm_cons Suc_length_conv is_perm_implies_same_length is_perm_refl nat.inject)
  qed
qed

(*
(* This not true, since reordering i and j can be prevented by an interfering element. *)
lemma standard_ordered:
  "0\<le>i \<Longrightarrow> i<j \<Longrightarrow> j<length S \<Longrightarrow>
   \<not> (standard S)!j \<prec> (standard S)!i"
oops
*)

(* Correctness, part II: standardization gives a non-decreasing list. *)

(* The actual lemma for correctness, part II *)
lemma standard_non_decreasing: "non_decreasing_list (standard S)"
proof(induct "length S" arbitrary:S)
  case 0 then show ?case by simp
next
  case (Suc n')
  note Suc' = Suc then show ?case
  proof(cases "n'")
    case 0 then show ?thesis using Suc'
         by (metis Suc_less_eq is_perm_implies_same_length less_nat_zero_code non_decreasing_list_def standard_is_perm)
  next
    case (Suc n)
    (* Helpers *)
    have S_length: "1 < length S"
         by (metis Suc Suc.hyps(2) diff_Suc_1 zero_less_Suc zero_less_diff)
    have remove1_rmpe: "Suc n = length (remove1 (rmpe S) S)"
         apply(simp add: length_remove1)
         by (metis Suc Suc.hyps(2) Zero_not_Suc diff_Suc_Suc diff_zero in_set_member length_0_conv rmpe_member)
    (* Applied induction hypothesis *)
    have non_decreasing_IH: "non_decreasing_list (standard (remove1 (rmpe S) S))" using Suc Suc'
         by(metis remove1_rmpe)
    (* Step *)
    have non_decreasing_step: "\<not> rmpe (remove1 (rmpe S) S) \<prec> rmpe S" using S_length
         by(simp add:rmpe_rmpe_remove1)
    (* The case *)
    show ?thesis
         apply(simp add:standard_not_nil_adhoc[OF Suc'(2)] standard_not_nil_adhoc[OF remove1_rmpe])
         by(metis non_decreasing_IH non_decreasing_step remove1_rmpe standard_not_nil_adhoc)
  qed
qed


(* Correctness, part III: standardization preserves the relative order of interfering elements. *)

lemma standard_preserves_interferences:
  "\<forall>S. \<exists>ip. is_perm [0 ..<length S] ip
         \<and> (standard S = map (op ! S) ip)
         \<and> (\<forall>k<length S. \<forall>i<k.
             (standard S)!i ## (standard S)!k
             \<longrightarrow> ip!i < ip!k)"
proof -
{ (* Fix a list [S] and state the goal. Proof will be by induction on the length of [S]. *)
  fix S
  have "\<exists>ip. is_perm [0 ..<length S] ip
         \<and> (standard S = map (op ! S) ip)
         \<and> (\<forall>k<length S. \<forall>i<k.
             (standard S)!i ## (standard S)!k
             \<longrightarrow> ip!i < ip!k)"
  proof(induct "length S" arbitrary: S)
    case 0 then show ?case by simp
  next
    case (Suc n)
    
    (* Let [r] be the rightmost permutable element of [S], that will be permuted at this step,
      write [i] its index in [S]. *)
    def r  \<equiv> "rmpe S"
    def i  \<equiv> "first_occ r S"
    have r_pe: "pe r S"
         by (metis Suc.hyps(2) length_Suc_conv neq_Nil_conv r_def rmpe_pe)
    have r_member: "List.member S r"
         by (metis pe_member r_pe)
    have i_range: "i < length S"
         by (metis first_occ_member_length i_def r_member)
    
    (* Write [T] for the list obtained by removing [r] from [S]. *)
    def T  \<equiv> "remove1 r S"
    have po_S: "standard S = r # (standard T)"
         by (metis local.pe_nil standard_not_nil T_def r_def r_pe)
    have T_length: "length S = Suc (length T)"
         by (metis Suc_length_conv is_perm_implies_same_length standard_is_perm po_S)
    have T_length': "Suc (length T) = length S"
         by (simp add:T_length)
    
    (* Write [U] for the list obtained by bringing [r] to the head of the list (but not yet
      reordering the tail [T]. *)
    def U  \<equiv> "r#T"
    have U_length: "length S = length U"
         by (metis T_length' U_def length_Suc_conv)
    
    (* By the crucial lemma, get the permutation [ipS] bringing [r] at the head of the list.
      This is a link between [S] and [U]. *)
    obtain ipS where ipS_cond: "is_perm [0 ..<length S] ipS
                               \<and> (U = map (op ! S) ipS)
                               \<and> (\<forall>k<length S. \<forall>j<k. U!j ## U!k \<longrightarrow> ipS!j < ipS!k)"
           using U_def T_def r_pe permuting_pe_preserves_interferences
           by metis
           (* Funny: this one is not found by sledgehammer ^_^ *)
    
    (* By induction hypothesis, get the permutation [ipT] reordering [T].
      This is a link between [T] and [standard T]. *)
    obtain ipT where ipT_cond: "is_perm [0 ..<length T] ipT
                               \<and> (standard T = map (op ! T) ipT)
                               \<and> (\<forall>k<length T. \<forall>i<k.
                                   (standard T)!i ## (standard T)!k
                                   \<longrightarrow> ipT!i < ipT!k)"
           using Suc T_length
           by (metis nat.inject)
    
    (* Define [ipU] as the extension of [ipT] that takes into account the new head [r] that
      has not to be moved again, then prove the same three properties as we have on [ipT].
      This is a link between [U = r#T] and [standard S = r#(standard T)]. *)
    def ipU \<equiv> "0 # map Suc ipT"
    have ipSC_cons1: "is_perm [1 ..<length S] (map Suc ipT)"
         apply (simp only:T_length upt_Suc_map)
         by (rule is_perm_map, simp_all add:ipT_cond)
    (* Property 1: [ipU] is a permutation of the indexes of [S]. *)
    have ipU_cond1: "is_perm [0 ..<length S] ipU"
         apply (simp add:ipU_def ipT_cond ipSC_cons1)
         by (metis One_nat_def Suc.hyps(2) ipSC_cons1 is_perm_skip upt_rec zero_less_Suc)
    have ipU_gtlS: "\<forall>a. List.member ipU a \<longrightarrow> a < length S"
         by (metis ipU_cond1 is_perm_member is_perm_sym member_range)
    (* Property 2: [ipU] relates [U] and [standard S]. *)
    have ipU_cond2: "standard S = map (op ! U) ipU"
         by (simp add:po_S ipT_cond ipU_def U_def)
    have ipU_cond2': "standard S = map (op ! (map (op ! S) ipS)) ipU"
         by (metis ipS_cond ipU_cond2)
    (* Property 3: [ipU] preserves the relative order of interfering elements. *)
    have ipU_cond3: "\<forall>k<length S. \<forall>i<k. (standard S)!i ## (standard S)!k \<longrightarrow> ipU!i < ipU!k"
    proof -
    { fix i k
      assume k_range: "k < length S"
         and i_range: "i < k"
         and ik_itrf: "(standard S)!i ## (standard S)!k"
      then have "ipU!i < ipU!k"
      proof(cases i)
        case 0 then show ?thesis using k_range i_range ik_itrf
             apply (simp add:ipU_def)
             by (metis One_nat_def Suc.hyps(2) Suc_diff_Suc Suc_less_eq ipSC_cons1 is_perm_implies_same_length length_map length_upt minus_nat.diff_0 nth_map zero_less_Suc)
      next
        case (Suc i')
        obtain k' where sk': "k = Suc k'" using i_range
               by (metis less_nat_zero_code not0_implies_Suc)
        then have "ipT ! i' < ipT ! k'" using Suc k_range i_range ik_itrf
             apply (simp add:ipU_def po_S)
             by (metis Suc_lessD T_length ipT_cond less_antisym)
        then show ?thesis
             apply (simp add:ipU_def Suc sk')
             by (metis One_nat_def Suc Suc.hyps(2) diff_Suc_Suc dual_order.strict_trans i_range ipSC_cons1 is_perm_implies_same_length k_range length_map length_upt minus_nat.diff_0 not_less_eq nth_map sk')
      qed
    } then show ?thesis by simp
    qed
    
    (* Define [ip], our candidate permutation, as the combination of [ipS] and [ipU]. *)
    def ip  \<equiv> "map (op ! ipS) ipU"
    
    (* First property: [ip] is a permutation of the indexes of [S]. *)
    have *: "is_perm [0..<length S] ip"
         apply (simp only:ip_def ipS_cond)
         by (metis ipS_cond ipU_cond1 is_perm_comp is_perm_implies_same_length is_perm_trans length_map)
    
    (* Second property: [ip] is the permutation relating [S] and [standard S]. *)
    have **: "standard S = map (op ! S) ip"
         apply (simp only:ip_def ipU_cond2 ipS_cond)
         by (metis ipS_cond ipU_cond1 is_perm_sym map_comp)
    
    (* Third property: [ip] preserves the relative order of interfering elements. *)
    { fix i k
      assume i_range: "i < k"
         and k_range: "k < length S"
         and ik_itrf: "standard S ! i ## standard S ! k"
      have i_range': "i < length S" using i_range k_range
           by (metis order.strict_trans)
      have ipU_ik_lt: "ipU ! i < ipU  ! k" using i_range k_range ik_itrf
           by (metis ipU_cond3)
      have ipU_k_lt: "ipU ! k < length S" using k_range
           by (metis T_length U_def bang_member ipS_cond ipU_cond1 ipU_gtlS is_perm_implies_same_length le0 length_Suc_conv length_map)
      have ipU_ik_itrf: "U ! (ipU ! i) ## U ! (ipU ! k)" using ik_itrf k_range i_range'
           by (metis ipU_cond2 is_perm_map_nth U_length ipU_cond1)
      have "ip ! i < ip ! k"
           by (metis k_range i_range' U_length ipS_cond ipU_ik_itrf ipU_ik_lt ipU_k_lt ip_def ipU_cond1 is_perm_map_nth length_map)
    } note *** = this
    
    (* Conclusion: [ip] is a valid candidate for our existential property. *)
    show ?case by (metis "*" "**" "***")
  qed
} 

(* Wrapping it up *)
then show ?thesis by simp
qed

value "take 2 [1,2,3]"

lemma is_perm_decopose : "  is_perm [0..< Suc n] ip 
         \<Longrightarrow> \<exists> ip' k. is_perm [1..< Suc n] ip' 
                      \<and> take  k ip @ drop (Suc k) ip = ip' 
                      \<and> k <= n 
                      \<and> ip ! k = 0"  
sorry

lemma mbind_eq_sexec: 
assumes * : "\<And>b \<sigma>'. f a \<sigma> = Some(b,\<sigma>') \<Longrightarrow> 
             (os \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S f; P (b#os)) = (os \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S f; P' (b#os))"
shows       "( a \<leftarrow> f a;  x \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S f; P (a # x)) \<sigma> = 
             ( a \<leftarrow> f a;  x \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S f; P' (a # x)) \<sigma>"
   apply(cases "f a \<sigma> = None")
         apply(subst bind_SE_def, simp)
         apply(subst bind_SE_def, simp)
        apply auto
         apply(subst bind_SE_def, simp)
         apply(subst bind_SE_def, simp)
        apply(simp add: *)
done

lemma mbind_eq_sexec': 
assumes * : "\<And>b \<sigma>'. f a \<sigma> = Some(b,\<sigma>') \<Longrightarrow> 
             (P (b))\<sigma>' = (P' (b))\<sigma>'"
shows       "( a \<leftarrow> f a;  P (a)) \<sigma> = 
             ( a \<leftarrow> f a;  P' (a)) \<sigma>"
   apply(cases "f a \<sigma> = None")
         apply(subst bind_SE_def, simp)
         apply(subst bind_SE_def, simp)
        apply auto
         apply(subst bind_SE_def, simp)
         apply(subst bind_SE_def, simp)
        apply(simp add: *)
   done

lemma mbindSF_compose:
"(os \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p (S@T) f; P os) = (os \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S f; os' \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p T f; P (os @ os'))"
proof (rule ext, rename_tac "\<sigma>", induct S arbitrary: \<sigma> P) 
   case Nil show ?case by simp
next
   case (Cons a S) show ?case 
        apply(insert Cons.hyps, simp)
        by(rule mbind_eq_sexec',simp)
qed

(* not used :
fun   rotR :: "'e list \<Rightarrow> 'e list"
where"rotR [] = []"
    |"rotR S  = (last S) # (butlast S)"
find_theorems "last" "butlast"

lemma test: "rotR [1::int,2,3]=[3,1,2]" (* why does value not work here ? *)
by simp

declare rotR.simps[simp del]

fun   rotL :: "'e list \<Rightarrow> 'e list"
where"rotL [] = []"
    |"rotL (a#S)  = S @ [a]"


lemma test': "rotL [1::int,2,3]=[2,3,1]" 
by simp
value "rotL[1::int,2,3]" (* why does value not work here ? *)

lemma rotL_charn : "rotL (S) = rotate1 S"
by(induct S, simp_all)

declare rotL.simps[simp del]

*)



lemma   action_commute : 
assumes *: "\<forall>x \<in> set S. \<not> x ## a"
shows     "(os \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p (S @ [a]) iostep ; (P' os)::'a \<Rightarrow> ('b \<times> 'a) option)  =  
           (os \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p (a # S)   iostep ; P' (rotate1 os))"
proof (rule ext, rename_tac "\<sigma>", rule rev_mp[OF *],induct S arbitrary: \<sigma> P') 
   case Nil show ?case by (simp )
next 
   case (Cons b S) 
   have * : "\<not> b ## a \<Longrightarrow>
             ( a \<leftarrow> iostep a;  
               aa \<leftarrow> iostep b;  
               os \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S iostep; 
               P' (rotate1 (a # aa # os))) =
             ( aa \<leftarrow> iostep b;  
               a \<leftarrow> iostep  a;  
               os \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S iostep; 
               P' (rotate1 (a # aa # os)))" 
             apply(insert interfer_consistent
                          [of "\<lambda> aa a . os \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S  iostep; P' (rotate1(a # aa # os))"])
             apply(simp add: itrf_is_exec_consistent_def)
             done
   have ** : "\<And> a aa os. rotate1 (a # aa # os) = aa # os @ [a]" by(simp)
   show ?case 
       apply(subst mbindSF_compose)
       apply(auto simp: *  simp del: rotate1.simps)
       apply(rule mbind_eq_sexec')
       by(subst Cons.hyps[simplified mbindSF_compose,simplified, THEN mp], simp_all add: **)
qed


lemma   action_commute2 : 
assumes *: "\<forall>x \<in> set S. \<not> x ## a"
shows     "(os \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p (S @ [a]) iostep ; 
            os' \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p T iostep ; 
            (P' (os @ os'))::'a \<Rightarrow> ('b \<times> 'a) option)  =  
           (os \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p (a # S)   iostep ; 
            os' \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p T iostep ; 
            P' ((rotate1 os)@os'))"
by(subst action_commute, simp_all add:*)

lemma   action_commute3 : 
assumes *: "\<forall>x \<in> set S. \<not> x ## a"
shows     "(os' \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p T iostep ; 
            os \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p (S @ [a]) iostep ; 
            (P' (os' @ os))::'a \<Rightarrow> ('b \<times> 'a) option)  =  
           (os' \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p T iostep ; 
            os \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p (a # S)   iostep ; 
            P' (os' @ (rotate1 os)))"
by(subst action_commute, simp_all add:*)


text{* the following definition of \emph{interference equivqlent executions}
serves as an intermediate step between the theory of permutations and 
their executions ... *}
definition itrf_equivalent :: "'b list \<Rightarrow> 'b list \<Rightarrow> bool"
where     "itrf_equivalent S S' = (\<exists>ip. is_perm [0 ..<length S] ip
                                    \<and> (S' = map (op ! S) ip)
                                    \<and> (\<forall>k<length S. \<forall>i<k. (S')!i ## (S')!k \<longrightarrow> ip!i < ip!k))"

declare List.upt.simps(2)[simp del]

lemma ordered_exec_state_invariant'' :
assumes * : "itrf_equivalent S S'"
shows "(\<sigma> \<Turnstile> (_ \<leftarrow> (mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S  iostep); assert\<^sub>S\<^sub>E (P))) = 
       (\<sigma> \<Turnstile> (_ \<leftarrow> (mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S' iostep); assert\<^sub>S\<^sub>E (P)))"

sorry


lemma ordered_exec_state_invariant' :
           "(\<sigma> \<Turnstile> (_ \<leftarrow> (mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p (S) iostep); assert\<^sub>S\<^sub>E (P))) = 
            (\<sigma> \<Turnstile> (_ \<leftarrow> (mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p (standard S) iostep); assert\<^sub>S\<^sub>E (P)))" 
sorry


find_theorems "SOME _._"

lemma ordered_exec_state_invariant_general :
   "(\<sigma> \<Turnstile> (os \<leftarrow> (mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S iostep); M S os)) =
            (let S' = reorder (\<lambda>(x,y). \<lambda>(x',y'). x \<preceq> x') (zip S [0 ..< length S]);
                 input_perm = map fst S';
                 index_perm = map snd S'
             in  \<sigma> \<Turnstile> (os \<leftarrow> (mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p input_perm iostep); 
                      M input_perm (map (op ! os) index_perm)))" 
sorry


end (* locale *)


end